import json
from typing import Callable


def do_function_call(function_call, ctx, tools):
    assert function_call["type"] == "function_call"
    assert isinstance(tools, list)
    for tool in tools:
        assert isinstance(tool, Callable)

    name = function_call["name"]
    arguments = function_call["arguments"]
    call_id = function_call["call_id"]

    # find the tool
    fn = None
    for tool in tools:
        if tool.__name__ == name:
            fn = tool
            break
    assert fn is not None

    # call
    args = json.loads(arguments)
    result = fn(ctx, **args)

    return {
        "type": "function_call_output",
        "call_id": call_id,
        "output": str(result),
    }


def use_tools(new_response_fn, messages, ctx, tools):
    messages = messages.copy()
    ret = None
    while True:
        res = new_response_fn(messages)
        pending = []
        for output in res.output:
            match output.type:
                case "function_call":
                    function_call = {
                        "type": "function_call",
                        "name": output.name,
                        "call_id": output.call_id,
                        "arguments": output.arguments,
                    }
                    pending.append(function_call)
                    messages.append(function_call)
                case "message":
                    ret = output.content[0].text
                case _:
                    assert False, f"unexpected output {output.type}"
        if len(pending) == 0:
            assert isinstance(ret, str) and len(ret) > 0
            return ret
        for function_call in pending:
            function_call_output = do_function_call(function_call, ctx, tools)
            messages.append(function_call_output)



def gen_obj(**props):
    """Generate schema for object. Properties are the fields of the object."""
    return {
        "type": "object",
        "properties": props,
        "required": list(props.keys()),
        "additionalProperties": False,
    }


def _wrap_arr(items):
    return {
        "type": "array",
        "items": items,
    }


def gen_arr(**props):
    """Generate schema for array of objects. Properties are the fields of the object."""
    return _wrap_arr(gen_obj(**props))


def gen_str(desc, enum=None, array=False):
    """Generate schema for string. When array is true, generate array of strings. Enum optionally specifies the allowed values."""
    assert isinstance(desc, str)
    assert enum is None or isinstance(enum, list)
    assert isinstance(array, bool)

    ret = {"type": "string", "description": desc}
    if enum is not None:
        ret["enum"] = enum
    if array:
        ret = _wrap_arr(ret)
    return ret


def gen_num(desc, array=False):
    """Generate schema for number. When array is true, generate array of numbers."""
    assert isinstance(desc, str)
    assert isinstance(array, bool)

    ret = {"type": "number", "description": desc}
    if array:
        ret = _wrap_arr(ret)
    return ret


def gen_bool(desc, array=False):
    """Generate schema for boolean. When array is true, generate array of booleans."""
    assert isinstance(desc, str)
    assert isinstance(array, bool)

    ret = {"type": "boolean", "description": desc}
    if array:
        ret = _wrap_arr(ret)
    return ret


def gen_schema(**props):
    """The top level schema which is always an object. Properties are the fields of the object."""
    return {
        "type": "json_schema",
        "json_schema": {
            "name": "output",
            "strict": True,
            "schema": gen_obj(**props),
        },
    }


def format_json_schema(**props):
    return {
        "format": {
            "type": "json_schema",
            "name": "output",
            "strict": True,
            "schema": gen_obj(**props),
        },
    }



from llmutil import gen, gen_schema, gen_str

sysmsg = "You are a helpful assistant that can answer questions and help with tasks."

result = gen(
    sysmsg,
    "What is the capital of the Japan?",
    gen_schema(
        answer=gen_str("Name of the capital"),
    ),
)
# {'answer': 'Tokyo'}
print(result)

result = gen(
    sysmsg,
    "What is the longest river in the world? A: Nile, B: Amazon, C: Yangtze, D: Mississippi",
    gen_schema(
        answer=gen_str("Choice", enum=["A", "B", "C", "D"]),
    ),
)
# {'answer': 'A'}
print(result)


result = gen(
    sysmsg,
    "what are the pokemon types?",
    gen_schema(
        types=gen_str("Pokemon types", array=True),
    ),
)
# {'types': ['Bug', 'Dark', 'Dragon', 'Electric', 'Fairy', 'Fighting', 'Fire', 'Flying', 'Ghost', 'Grass', 'Ground', 'Ice', 'Normal', 'Poison', 'Psychic', 'Rock', 'Steel', 'Water']}
print(result)



from llmutil import chat

output = chat(
    [
        {"role": "system", "content": "You are a helpful assistant."},
        {"role": "user", "content": "tell me a joke"},
    ],
    model="gpt-4o-mini",
)
print(output)



import json

from openai import OpenAI

from .client import default_client


def chat(messages, model="gpt-4o", client=None, **kwargs):
    assert isinstance(messages, list) and len(messages) > 0
    assert isinstance(model, str)
    assert isinstance(client, OpenAI) or client is None

    if client is None:
        client = default_client()

    response = client.chat.completions.create(
        model=model,
        messages=messages,
        **kwargs,
    )
    return response.choices[0].message.content


def gen(sysmsg, usrmsg, response_format, model="gpt-4o", client=None, **kwargs):
    assert isinstance(sysmsg, str)
    assert isinstance(usrmsg, str)
    assert isinstance(response_format, dict)
    assert isinstance(model, str)
    assert isinstance(client, OpenAI) or client is None

    if client is None:
        client = OpenAI()

    messages = [
        {"role": "system", "content": sysmsg},
        {"role": "user", "content": usrmsg},
    ]
    response = client.beta.chat.completions.parse(
        model=model,
        messages=messages,
        response_format=response_format,
        **kwargs,
    )
    return json.loads(response.choices[0].message.content)


def ask(instructions, question):
    response = default_client().responses.create(
        model="gpt-4o",
        input=[
            {"role": "system", "content": instructions},
            {"role": "user", "content": question},
        ],
    )
    return response.output_text



import inspect
import json
from collections.abc import Callable

from .client import default_client, default_model


def tool_def(fn):
    assert isinstance(fn, Callable)

    name = fn.__name__
    doc = fn.__doc__
    assert doc is not None
    params = list(inspect.signature(fn).parameters.keys())

    instructions = f"""generate a tool definition for this function.
name: {name}
docstring: {doc}
params: {params}"""

    text = {
        "format": {
            "type": "json_schema",
            "name": "tool_def",
            "schema": {
                "type": "object",
                "properties": {
                    "description": {
                        "type": "string",
                        "description": "description of the tool",
                    },
                    "params": {
                        "type": "object",
                        "description": "parameters of the tool",
                        "properties": {
                            p: {
                                "type": "object",
                                "properties": {
                                    "type": {
                                        "type": "string",
                                        "enum": ["string", "number"],
                                        "description": "type of the parameter, can only be string or number",
                                    },
                                    "enum": {
                                        "type": ["array", "null"],
                                        "items": {
                                            "type": "string",
                                        },
                                        "description": "optional, allowed values of the parameter, only applicable when type is string",
                                    },
                                    "description": {
                                        "type": "string",
                                        "description": "description of the parameter",
                                    },
                                },
                                "additionalProperties": False,
                                "required": ["type", "enum", "description"],
                            }
                            for p in params
                        },
                        "additionalProperties": False,
                        "required": params,
                    },
                },
                "required": ["description", "params"],
                "additionalProperties": False,
            },
            "strict": True,
        },
    }
    result = default_client().responses.create(
        model=default_model,
        input=instructions,
        text=text,
    )
    data = json.loads(result.output_text)

    output_params = {}
    for p in params:
        output_params[p] = {
            "type": data["params"][p]["type"],
            "description": data["params"][p]["description"],
        }
        if data["params"][p]["enum"] is not None:
            output_params[p]["enum"] = data["params"][p]["enum"]

    return {
        "type": "function",
        "name": name,
        "description": data["description"],
        "strict": True,
        "parameters": {
            "type": "object",
            "properties": output_params,
            "required": params,
            "additionalProperties": False,
        },
    }



from .llm import ask, chat, gen
from .schema import (
    format_json_schema,
    gen_arr,
    gen_bool,
    gen_num,
    gen_obj,
    gen_schema,
    gen_str,
)
from .tools import use_tools
from .tools_def import tool_def

__all__ = [
    "chat",
    "gen",
    "ask",
    "gen_arr",
    "gen_bool",
    "gen_num",
    "gen_obj",
    "gen_schema",
    "format_json_schema",
    "gen_str",
    "tool_def",
    "use_tools",
]
